/*
    learn.c
    Unigram Mixtures, simple learning procedure.
    $Id: learn.c,v 1.5 2013/01/09 05:35:03 daichi Exp $

*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "learn.h"
#include "likelihood.h"
#include "feature.h"
#include "dmatrix.h"
#include "random.h"
#include "util.h"

void
um_learn (document *data, double *lambda, double **beta,
	  int nmixtures, int nlex, double gamma, double eta,
	  int emmax, double epsilon)
{
	document *dp;
	double **p;
	double z, ppl, pppl;
	int i, j, k, n, t, v;

	/* initialize seed */
	srand(time(NULL));

	/* initialize lambda */
	for (k = 0; k < nmixtures; k++)
		lambda[k] = 1.0 / (double)nmixtures;

	/* initialize beta */
	for (k = 0; k < nmixtures; k++)
	{
		z = 0;
		for (v = 0; v < nlex; v++) {
			beta[k][v] = RANDOM + 10.0;
			z += beta[k][v];
		}
		for (v = 0; v < nlex; v++) {
			beta[k][v] /= z;
		}
	}
	
	/* count data length, allocate p */
	for (dp = data, n = 0; (dp->len) != -1; dp++, n++)
		;
	if ((p = dmatrix(n, nmixtures)) == NULL) {
		fprintf(stderr, "um_learn:: can't allocate p.\n");
		return;
	}

	for (t = 0; t < emmax; t++)
	{
		/*
		 * E step
		 *
		 */
		for (dp = data, i = 0; (dp->len) != -1; dp++, i++)
		{
			if (i % 100 == 0)
			{
				printf("iteration %2d/%2d.. (%4d/%4d)\r",
				       t + 1, emmax, i, n);
				fflush(stdout);
			}
			
			for (k = 0; k < nmixtures; k++)
			{
				for (j = 0, z = 0; j < dp->len; j++) {
					z += dp->cnt[j] * log (beta[k][dp->id[j]]);
				}
				p[i][k] = log (lambda[k]) + z;
			}
			/* normalize, and exp */
			for (k = 0, z = 0; k < nmixtures; k++)
				z = logsumexp(z, p[i][k], (k == 0));
			for (k = 0; k < nmixtures; k++) {
				p[i][k] = exp(p[i][k] - z);
			}
		}
		printf("iteration %2d/%2d.. (%2d/%2d)\t",
		       t + 1, emmax, i, n); fflush(stdout);
		
		/*
		 * M step
		 *
		 */
		/* initialize lambda */
		for (k = 0; k < nmixtures; k++)
			lambda[k] = gamma;
		for (dp = data, i = 0; (dp->len) != -1; dp++, i++)
		{
			for (k = 0; k < nmixtures; k++)
				lambda[k] += p[i][k];
		}
		/* normalize */
		for (k = 0, z = 0; k < nmixtures; k++)
			z += lambda[k];
		for (k = 0; k < nmixtures; k++)
			lambda[k] /= z;
		
		/* initialize beta */
		for (k = 0; k < nmixtures; k++)
			for (v = 0; v < nlex; v++)
				beta[k][v] = eta;
		/* accumulate */
		for (dp = data, i = 0; (dp->len) != -1; dp++, i++)
		{
			for (k = 0; k < nmixtures; k++)
			{
				for (j = 0; j < dp->len; j++)
					beta[k][dp->id[j]] += p[i][k] * dp->cnt[j];
			}
		}
		/* normalize for expectation */
		for (k = 0; k < nmixtures; k++)
		{
			for (v = 0, z = 0; v < nlex; v++)
				z += beta[k][v];
			for (v = 0; v < nlex; v++)
				beta[k][v] /= z;
		}
		/*
		 *  converge?
		 *
		 */
		ppl = um_ppl (data, p, lambda, beta, nmixtures);
		printf("PPL = %lg\n", ppl);
		
		if (t > 1 && converged(ppl, pppl, epsilon))
		{
			printf("converged.\n");
			break;
		} else {
			pppl = ppl;
			printf("\r");
		}
	}
	
}

